<?php

/**
 * Copyright 2017-2024 IQRF Tech s.r.o.
 * Copyright 2019-2024 MICRORISC s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
declare(strict_types = 1);

namespace Iqrf\CommandExecutor\DI;

use Iqrf\CommandExecutor\CommandExecutor;
use Iqrf\CommandExecutor\CommandStack;
use Nette\DI\CompilerExtension;
use Nette\PhpGenerator\ClassType;
use Nette\Schema\Expect;
use Nette\Schema\Schema;
use stdClass;

/**
 * Nette DI extension for command executor
 * @property stdClass $config Configuration
 */
class CommandExecutorExtension extends CompilerExtension {

	/**
	 * Returns the configuration schema
	 * @return Schema Configuration schema
	 */
	public function getConfigSchema(): Schema {
		return Expect::structure([
			'debug' => Expect::bool()
				->default(false)
				->dynamic(),
			'sudo' => Expect::bool()
				->default(true)
				->dynamic(),
		]);
	}

	/**
	 * Loads extension configuration
	 */
	public function loadConfiguration(): void {
		$builder = $this->getContainerBuilder();

		/// Command stack registration
		$builder->addDefinition($this->prefix('commandStack'))
			->setType(CommandStack::class);

		/// Command executor registration
		$builder->addDefinition($this->prefix('commandExecutor'))
			->setType(CommandExecutor::class)
			->setArguments([
				$this->config->sudo,
				'@' . $this->prefix('commandStack'),
			]);
	}

	/**
	 * Adjusts DI container compiled to PHP class - adds Tracy panel
	 * @param ClassType $class DI container compiled to PHP class
	 */
	public function afterCompile(ClassType $class): void {
		if ($this->config->debug) {
			$initialize = $class->getMethod('initialize');
			$initialize->addBody(
				'$this->getService(?)->addPanel(new \Iqrf\CommandExecutor\Tracy\CommandPanel($this->getService(?)));',
				['tracy.bar', $this->prefix('commandStack')]
			);
		}
	}

}
