<?php
/**
 * Copyright 2018 IQRF Tech s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
declare(strict_types = 1);

namespace Iqrf\IdeMacros;

/**
 * IQRF IDE Macro file parser
 */
class MacroFileParser {

	/**
	 * @var int IQRF IDE Macro's file version
	 */
	public $version;

	/**
	 * @var string IQRF IDE Macro's file description
	 */
	public $description;

	/**
	 * @var int IQRF IDE Macro's file type
	 */
	public $type;

	/**
	 * @var Group[] IQRF IDE Macro's groups
	 */
	public $groups;

	/**
	 * @var string IQRF IDE Macro's file name
	 */
	private $fileName;

	public function __construct(string $fileName) {
		$this->fileName = $fileName;
	}

	/**
	 * Read IQRF IDE Macro's file and parse it
	 * @return Group[] Parsed macros
	 */
	public function read(): array {
		$fileContent = parse_ini_file($this->fileName, true)['Macro'];
		$this->version = intval($fileContent['Version']);
		$this->description = $fileContent['Description'];
		$this->type = intval($fileContent['Type']);
		return $this->parse($fileContent);
	}

	/**
	 * Parse IQRF IDE Macros into objects
	 * @param mixed[] $fileContent IQRF IDE Macro's file content
	 * @return Group[] Array of objects with IQRF IDE macros
	 */
	private function parse(array $fileContent): array {
		$ascii = Hex::toAscii($fileContent['Macros']);
		$array = array_slice(explode("\r\n", trim($ascii)), 3);
		$chunks = array_chunk($array, 12 * $this->getMacroSize() + 3);
		$groups = [];
		foreach ($chunks as $chunk) {
			$groups[] = new Group($chunk, $this);
		}
		return $groups;
	}

	/**
	 * Returns IQRF IDE Macro's size
	 * @return int IQRF IDE Macro's size
	 */
	public function getMacroSize(): int {
		if ($this->version === 262147) {
			$macroSize = 6;
		} else {
			$macroSize = 5;
		}
		return $macroSize;
	}

	/**
	 * Convert objects into arrays
	 * @param mixed[] $macros IQRF IDE Macros in an array of objects
	 * @return mixed[] IQRF IDE Macros in an array
	 */
	public function toArray(array $macros): array {
		$array = [];
		foreach ($macros as $id => $group) {
			assert($group instanceof Group);
			$array[$id] = $group->toArray();
		}
		return $array;
	}

}
