<?php

/**
 * Copyright 2019-2024 IQRF Tech s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
declare(strict_types = 1);

namespace Iqrf\Repository\Models;

use Iqrf\Repository\Entities\Company;
use Iqrf\Repository\Exceptions\CompanyNotFound;
use Iqrf\Repository\Exceptions\CorruptedData;
use Iqrf\Repository\Exceptions\DataNotFound;
use Iqrf\Repository\Exceptions\ServiceUnavailable;

/**
 * Company manager
 */
final class CompanyManager extends BaseManager {

	/**
	 * Companies path
	 */
	private const PATH = 'companies';

	/**
	 * Lists all companies
	 * @return array<Company> All companies
	 * @throws CorruptedData Corrupted data in the API response
	 * @throws ServiceUnavailable IQRF Repository API is unavailable
	 */
	public function list(): array {
		return array_map(
			static fn (array $response): Company => Company::fromApiResponse($response),
			$this->apiClient->sendApiRequest(self::PATH),
		);
	}

	/**
	 * Returns the company by its ID
	 * @param int $id Company ID
	 * @return Company Company
	 * @throws CorruptedData Corrupted data in the API response
	 * @throws CompanyNotFound Requested company not found
	 * @throws ServiceUnavailable IQRF Repository API is unavailable
	 */
	public function get(int $id): Company {
		try {
			$response = $this->apiClient->sendApiRequest(self::PATH . '/' . $id);
		} catch (DataNotFound) {
			throw new CompanyNotFound();
		}
		return Company::fromApiResponse($response);
	}

}
